@@ -23,8 +23,11 @@ from xml.sax.saxutils import unescape
 from kiwi.ui import dialogs
 from kiwi.environ import environ
 
+from gazpacho.config import get_app_data_dir
 from gazpacho.library import load_library, LibraryLoadError
-from gazpacho.loader import tags
+from gazpacho.loader.custom import valuefromstringsimpletypes, \
+                                    get_pspec_from_name, str2bool
+from gazpacho.properties import prop_registry
 from gazpacho.widgetregistry import widget_registry
 
 def xml_get_text_from_node(xmlnode):
@@ -83,7 +86,7 @@ class Catalog:
         # <!ELEMENT glade-catalog (glade-widget-classes?,
         #                          glade-widget-group*)>
 
-        if node.nodeName == tags.GLADE_CATALOG:
+        if node.nodeName == 'glade-catalog':
             self._parse_catalog(node)
         else:
             msg = "Root node is not glade catalog: %s" % node.nodeName
@@ -93,9 +96,9 @@ class Catalog:
             if child.nodeType != child.ELEMENT_NODE:
                 continue
 
-            if child.nodeName == tags.GLADE_WIDGET_CLASSES:
+            if child.nodeName == 'glade-widget-classes':
                 self._parse_widget_classes(child)
-            elif child.nodeName == tags.GLADE_WIDGET_GROUP:
+            elif child.nodeName == 'glade-widget-group':
                 self._parse_widget_group(child)
             else:
                 print 'Unknown node: %s' % child.nodeName
@@ -110,21 +113,21 @@ class Catalog:
         #                         requires        CDATA #IMPLIED>
 
         # name
-        self.title = str(node.getAttribute(tags.NAME))
+        self.title = str(node.getAttribute('name'))
 
         # priority
         priority = sys.maxint
-        if node.hasAttribute(tags.PRIORITY):
-            priority = int(node.getAttribute(tags.PRIORITY))
+        if node.hasAttribute('priority'):
+            priority = int(node.getAttribute('priority'))
         self.priority = priority
 
         # library
-        library_name = str(node.getAttribute(tags.LIBRARY))
+        library_name = str(node.getAttribute('library'))
 
         # library-type
         library_type = "python"
-        if node.hasAttribute(tags.LIBRARY_TYPE):
-            library_type = str(node.getAttribute(tags.LIBRARY_TYPE))
+        if node.hasAttribute('library-type'):
+            library_type = str(node.getAttribute('library-type'))
 
         name = os.path.basename(self.filename[:-4])
         try:
@@ -137,7 +140,7 @@ class Catalog:
         "<!ELEMENT glade-widget-classes (glade-widget-class+)>"
 
         for class_node in node.childNodes:
-            if class_node.nodeName != tags.GLADE_WIDGET_CLASS:
+            if class_node.nodeName != 'glade-widget-class':
                 continue
 
             self._parse_widget_class(class_node)
@@ -150,28 +153,28 @@ class Catalog:
         #                              adaptor-class CDATA #IMPLIED>
         #                              register-type-function CDATA #IMPLIED>
 
-        name = str(node.getAttribute(tags.NAME))
+        name = str(node.getAttribute('name'))
         if widget_registry.has_name(name):
             raise CatalogError("WidgetAdaptor %s is already registered" % name)
 
-        generic_name = str(node.getAttribute(tags.GENERIC_NAME))
-        palette_name = str(node.getAttribute(tags.TITLE))
-        adaptor_class_name = str(node.getAttribute(tags.ADAPTOR_CLASS))
+        generic_name = str(node.getAttribute('generic-name'))
+        palette_name = str(node.getAttribute('title'))
+        adaptor_class_name = str(node.getAttribute('adaptor-class'))
 
         tooltip = None
         for child in node.childNodes:
-            if child.nodeName == tags.TOOLTIP:
+            if child.nodeName == 'tooltip':
                 tooltip = xml_get_text_from_node(child)
 
         try:
-            adaptor = self.library.get_widget_adaptor_class(name,
-                                                            adaptor_class_name)
+            adaptor_class = self.library.get_widget_adaptor_class(
+                name, adaptor_class_name)
         except LibraryLoadError, e:
             print 'Could not load %s: %s' % (name, e)
             return
 
-        adaptor.setup(name, generic_name, palette_name, self.library,
-                      self.resource_path, tooltip)
+        adaptor = adaptor_class(name, generic_name, palette_name, self.library,
+                                self.resource_path, tooltip)
 
         # store the Widget Adaptor on the cache
         widget_registry.add(adaptor)
@@ -181,16 +184,16 @@ class Catalog:
         #  <!ATTLIST glade-widget-group name  CDATA #REQUIRED
         #                               title CDATA #REQUIRED>
 
-        name = str(node.getAttribute(tags.NAME))
-        title = str(node.getAttribute(tags.TITLE))
+        name = str(node.getAttribute('name'))
+        title = str(node.getAttribute('title'))
 
         widget_group = WidgetGroup(name, title)
 
         for child in node.childNodes:
-            if child.nodeName != tags.GLADE_WIDGET_CLASS_REF:
+            if child.nodeName != 'glade-widget-class-ref':
                 continue
 
-            name = str(child.getAttribute(tags.NAME))
+            name = str(child.getAttribute('name'))
             widget_adaptor = widget_registry.get_by_name(name)
             if widget_adaptor is not None:
                 widget_group.append(widget_adaptor)
@@ -227,7 +230,7 @@ def _get_catalog_dirs():
     catalog_error = ""
 
     _parse_gazpacho_path()
-    retval = []
+    catalogs = []
     for dirname in environ.get_resource_paths('catalog'):
         try:
             dir_list = os.listdir(dirname)
@@ -247,9 +250,22 @@ def _get_catalog_dirs():
         for filename in dir_list:
             if not filename.endswith('.xml'):
                 continue
-            retval.append((filename[:-4], os.path.join(dirname, filename)))
+            name = filename[:-4]
+            item = name, os.path.join(dirname, filename)
+            # Make sure the base catalog is always loaded first, see #343283
+            # We should introduce proper dependencies in the future.
+            if name == 'base':
+                catalogs.insert(0, item)
+            else:
+                catalogs.append(item)
 
-    return retval
+    for name, filename in catalogs:
+        if name == 'base':
+            break
+    else:
+        raise SystemExit("Unable to find GTK+ catalog which is mandatory")
+
+    return catalogs
 
 _catalogs = []
 
@@ -270,21 +286,19 @@ def get_all_catalogs():
     return _catalogs
 
 def load_catalogs(external=True):
+    """
+    @param external: If true, allow loading of external (eg, not base) catalogs
+    """
     try:
         catalogs = _get_catalog_dirs()
     except EnvironmentError, e:
         dialogs.error("<b>There were errors while loading Gazpacho</b>", str(e))
         raise SystemExit
 
-    for name, filename in catalogs:
-        if name == 'base':
-            break
-    else:
-        raise SystemExit("Unable to find GTK+ catalog which is mandatory")
-
     problems = []
     for name, filename in catalogs:
-        if not external and os.path.basename(filename) != 'base.xml':
+        # Skip external catalogs, for the testsuite.
+        if not external and name != 'base':
             continue
 
         if filename.endswith('gtk+.xml'):
@@ -308,3 +322,143 @@ def load_catalogs(external=True):
 
     _catalogs.sort(lambda c1, c2: cmp(c1.priority,
                                       c2.priority))
+
+# Default name of the file that contains custom UI definitions
+CUSTOM_UI_FILENAME = 'custom.ui'
+
+def get_custom_ui_filename():
+    """Return the name of the file with custom UI definitions if it exists
+
+    First it look in the user home directory (in the .gazpacho subdirectory),
+    then it looks in every directory where there are catalogs.
+    """
+    # First we look in the user home directory
+    app_dir = get_app_data_dir('gazpacho')
+    custom_ui_filename = os.path.join(app_dir, CUSTOM_UI_FILENAME)
+    if os.path.exists(custom_ui_filename):
+        return custom_ui_filename
+
+    # Then, look at standard catalog places
+    for dirname in environ.get_resource_paths('catalog'):
+        custom_ui_filename = os.path.join(dirname, CUSTOM_UI_FILENAME)
+        if os.path.exists(custom_ui_filename):
+            return custom_ui_filename
+
+class CustomUI(object):
+    """A Custom User Interface definition
+
+    Using a XML file it is possible to redefine some parts of Gazpacho UI.
+    Right now you can:
+
+    - Change the palette grouping of widgets (you can mix widgets from
+      different catalogs into the same palette groups)
+
+    - Change the default value of a property of a widget
+
+    - Hide a property making it non editable
+
+    - Change the order of the properties in the editor by modifing its priority
+    """
+
+    def __init__(self, filename):
+        """Constructor.
+
+        It parses the file and changes the Gazpacho UI using its definitions
+        """
+        self.filename = filename
+        self.widget_groups = []
+
+        document = xml.dom.minidom.parse(file(filename))
+        root = document.documentElement
+
+        if root.nodeName != 'gazpacho-custom-ui':
+            msg = "Root node is not gazpacho-custom-ui: %s" % root.nodeName
+            raise BadCatalogSyntax(filename, msg)
+
+        self._parse_root(root)
+
+    def _parse_root(self, root):
+        """Parser start point"""
+        for child in root.childNodes:
+            if child.nodeType != child.ELEMENT_NODE:
+                continue
+
+            if child.nodeName == 'palette-group':
+                self._parse_custom_widget_group(child)
+            else:
+                print 'Unknown node: %s' % child.nodeName
+
+    def _parse_custom_widget_group(self, node):
+        """Parse a widget group and store it on a instance attribute"""
+        name = str(node.getAttribute('name'))
+        title = str(node.getAttribute('title'))
+
+        widget_group = WidgetGroup(name, title)
+
+        for child in node.childNodes:
+            if child.nodeType != child.ELEMENT_NODE:
+                continue
+
+            if child.nodeName == 'widget':
+                name = str(child.getAttribute('name'))
+                widget_adaptor = widget_registry.get_by_name(name)
+                if widget_adaptor is not None:
+                    widget_group.append(widget_adaptor)
+                    self._parse_custom_widget(name, child)
+                else:
+                    print 'Could not load widget class %s' % name
+
+            else:
+                print 'Unknown node: %s' % child.nodeName
+        self.widget_groups.append(widget_group)
+
+    def _parse_custom_widget(self, widget_name, node):
+        """Parse a widget changing its property types"""
+        for child in node.childNodes:
+            if child.nodeType != child.ELEMENT_NODE:
+                continue
+
+            if child.nodeName == 'property':
+                prop_name = str(child.getAttribute('name'))
+                full_name = '%s::%s' % (widget_name, prop_name)
+                attrs = {}
+                for i in range(child.attributes.length):
+                    attr = child.attributes.item(i)
+                    name = str(attr.name)
+                    value = attr.value
+                    if name == 'name':
+                        continue
+
+                    if name == 'default':
+                        value = _valuefromstring(widget_name, prop_name, value)
+
+                    elif name in ('readable', 'writable', 'enabled',
+                                  'translatable', 'editable'):
+                        value = str2bool(value)
+
+                    elif name == 'priority':
+                        value = int(value)
+
+                    else:
+                        print 'Customization of attribute %s is not supported' % name
+                        continue
+
+                    attrs[str(attr.name)] = value
+
+                if 'default' in attrs.keys():
+                    attrs['has_custom_default'] = True
+
+                prop_registry.override_simple(full_name, **attrs)
+
+            else:
+                print 'Unknown node: %s' % child.nodeName
+
+def _valuefromstring(widget_name, prop_name, string_value):
+    """Helper function to convert a string to a valid value
+
+    This is used when parsing default values for custom ui properties.
+    """
+    pspec = get_pspec_from_name(widget_name, prop_name)
+    prop_type = pspec.value_type
+    value = valuefromstringsimpletypes(prop_type, string_value)
+    return value
